import contextlib
from typing import List

from ..config import JointManagerDynamixelSerialConfig, M5StackSerialConfig
from ..joint_manager import JointManager
from ..m5stack_client import M5StackClient
from .dynamixel import DynamixelController
from .dynamixel_communicator import DynamixelCommunicator
from .m5stack import M5StackSerialClient
from .m5stack_communicator import M5SerialCommunicator


def create_joint_manager(
    config: JointManagerDynamixelSerialConfig, stack: contextlib.ExitStack
) -> JointManager:
    communicator = stack.enter_context(
        DynamixelCommunicator.open(
            serial_port=config.serial_port,
            baudrate=config.baudrate,
            protocol_version=config.protocol_version,
        )
    )
    controllers: List[DynamixelController] = []
    for c in config.controllers:
        controllers.append(
            DynamixelController(
                c.joint_name,
                c.dynamixel_id,
                communicator,
            )
        )

    return JointManager(controllers)


def create_m5stack_client(
    config: M5StackSerialConfig, stack: contextlib.ExitStack
) -> M5StackClient:
    communicator = stack.enter_context(
        M5SerialCommunicator(
            baudrate=config.baudrate,
            port=config.serial_port,
            timeout=config.serial_timeout,
        )
    )
    return M5StackSerialClient(communicator)



import contextlib
import logging
import threading
from typing import Dict

import grpc

_logger = logging.getLogger(__name__)
_lock = threading.Lock()
_channels: Dict[str, grpc.Channel] = {}


def create_channel(stack: contextlib.ExitStack, endpoint: str) -> grpc.Channel:
    with _lock:
        c = _channels.get(endpoint)
        if c is not None:
            _logger.debug(f"Use existing channel for endpoint: '{endpoint}'")
            return c

        channel = stack.enter_context(grpc.insecure_channel(endpoint))
        _logger.debug(f"New channel for endpoint: '{endpoint}' created")
        _channels[endpoint] = channel
        return channel



import contextlib
import logging
from typing import List

from ..config import JointManagerGrpcConfig, M5StackGrpcConfig
from ..joint_manager import JointManager
from .channel_pool import create_channel
from .joints_controller import GrpcJointController, _GrpcJointsController
from .m5stack import GrpcM5StackClient

_logger = logging.getLogger(__name__)


def create_joint_manager(
    config: JointManagerGrpcConfig, stack: contextlib.ExitStack
) -> JointManager:
    channel = create_channel(stack, config.endpoint)
    service_client = _GrpcJointsController(channel)
    service_joints = service_client.get_joint_names()
    _logger.debug(f"Server response: service_joints='{service_joints}'")
    controllers: List[GrpcJointController] = []

    target_joints = config.joints or service_joints
    for joint_name in target_joints:
        if joint_name not in service_joints:
            raise ValueError(
                f"joint_name: '{joint_name}' is not supported in "
                f"gRPC server: '{config.endpoint}'"
            )

        _logger.debug(f"Use grpc controller for joint: '{joint_name}'")
        controllers.append(GrpcJointController(joint_name, service_client))

    return JointManager(controllers)


def create_m5stack_client(
    config: M5StackGrpcConfig, stack: contextlib.ExitStack
) -> GrpcM5StackClient:
    channel = create_channel(stack, config.endpoint)
    return GrpcM5StackClient(channel)



import contextlib
import logging
import pathlib
from typing import List, Literal, Optional, Union

import pydantic

from .joint_manager import JointManager
from .m5stack_client import M5StackClient

_logger = logging.getLogger(__name__)


class DynamixelControllerConfig(pydantic.BaseModel):
    joint_name: str
    dynamixel_id: int
    min_position_limit: float
    max_position_limit: float
    default_velocity: float
    default_acceleration: float


class JointManagerDynamixelSerialConfig(pydantic.BaseModel):
    type: Literal["dynamixel_serial"]
    controllers: List[DynamixelControllerConfig] = pydantic.Field(default=[])
    serial_port: pathlib.Path = pathlib.Path("/dev/ttyUSB_dynamixel")
    baudrate: int = 1000000
    protocol_version: float = 2.0

    def factory(self, stack: contextlib.ExitStack) -> JointManager:
        from .serial.factory import create_joint_manager

        _logger.debug("Initializing joint manager from 'dynamixel_serial' config")
        return create_joint_manager(self, stack)


class JointManagerGrpcConfig(pydantic.BaseModel):
    type: Literal["grpc"]
    endpoint: str
    joints: Optional[List[str]]

    def factory(self, stack: contextlib.ExitStack) -> JointManager:
        from .grpc.factory import create_joint_manager

        _logger.debug("Initializing joint manager from 'grpc' config")
        return create_joint_manager(self, stack)


JointManagerConfig = Union[JointManagerDynamixelSerialConfig, JointManagerGrpcConfig]


class M5StackSerialConfig(pydantic.BaseModel):
    type: Literal["serial"]
    serial_port: pathlib.Path = pathlib.Path("/dev/ttyUSB_M5Stack")
    baudrate: int = 500000
    serial_timeout: float = 0.2

    def factory(self, stack: contextlib.ExitStack) -> M5StackClient:
        from .serial.factory import create_m5stack_client

        _logger.debug("Initializing M5Stack client from 'serial' config")
        return create_m5stack_client(self, stack)


class M5StackGrpcConfig(pydantic.BaseModel):
    type: Literal["grpc"]
    endpoint: str

    def factory(self, stack: contextlib.ExitStack) -> M5StackClient:
        from .grpc.factory import create_m5stack_client

        _logger.debug("Initializing M5Stack client from 'grpc' config")
        return create_m5stack_client(self, stack)


M5StackConfig = Union[M5StackSerialConfig, M5StackGrpcConfig]


class AkariClientConfig(pydantic.BaseModel):
    joint_manager: JointManagerConfig = pydantic.Field(..., discriminator="type")
    m5stack: M5StackConfig = pydantic.Field(..., discriminator="type")


class AkariClientEnv(pydantic.BaseSettings):
    config_path: Optional[pydantic.FilePath] = None

    class Config:
        env_prefix = "AKARI_CLIENT_"
        case_sensitive = False


_CONFIG_LOOKUP_PATHS = [
    pathlib.Path("~/.config/akari/client_config.json"),
    pathlib.Path("/etc/akari/client_config.json"),
]


def load_env() -> AkariClientEnv:
    return AkariClientEnv()


def _load_config(path: pathlib.Path) -> AkariClientConfig:
    _logger.debug(f"Load config: {path}")
    return AkariClientConfig.parse_file(path)


def default_serial_config() -> AkariClientConfig:
    return AkariClientConfig(
        joint_manager=JointManagerDynamixelSerialConfig(type="dynamixel_serial"),
        m5stack=M5StackSerialConfig(type="serial"),
    )


def load_config() -> AkariClientConfig:
    env = load_env()
    if env.config_path is not None:
        _logger.debug("env.config_path is set")
        return _load_config(env.config_path)

    for path in _CONFIG_LOOKUP_PATHS:
        path = path.expanduser().resolve()
        if path.exists():
            return _load_config(path)

    _logger.debug("Use default config")
    return default_serial_config()



from __future__ import annotations

import dataclasses
import enum


@dataclasses.dataclass(frozen=True)
class Color:
    """M5のディスプレイの表示色を指定するクラス。

    :param red: RGBのR要素。0~255で指定。
    :param green: RGBのG要素。0~255で指定。
    :param blue: RGBのB要素。0~255で指定。

    """

    red: int
    green: int
    blue: int

    def __post_init__(self) -> None:
        assert 0 <= self.red <= 0xFF
        assert 0 <= self.green <= 0xFF
        assert 0 <= self.blue <= 0xFF

    @staticmethod
    def from_rgb565(value: int) -> Color:
        # NOTE: Colors in RGB565 are represented as follows
        # value = (r7 r6 r5 r4 r3 g7 g6 g5 g4 g3 g2 b7 b6 b5 b4 b3)
        assert 0 <= value <= 0xFFFF
        blue = (value & 0x1F) << 3
        # lshift 5 bits and rshift 2 bits
        green = (value & 0x7E0) >> 3
        # lshift 11 bits and rshift 3 bits
        red = (value & 0xF800) >> 8

        return Color(red, green, blue)

    def as_rgb565(self) -> int:
        return ((self.red >> 3) << 11) | ((self.green >> 2) << 5) | (self.blue >> 3)


class Colors(Color, enum.Enum):
    """
    | M5のディスプレイの表示色のenum。
    | 下記の色を指定可能。
    | BLACK
    | NAVY
    | DARKGREEN
    | DARKCYAN
    | MAROON
    | PURPLE
    | OLIVE
    | LIGHTGREY
    | DARKGREY
    | BLUE
    | GREEN
    | CYAN
    | RED
    | MAGENTA
    | YELLOW
    | WHITE
    | ORANGE
    | GREENYELLOW
    | PINK

    """

    BLACK = (0, 0, 0)
    NAVY = (0, 0, 127)
    DARKGREEN = (0, 127, 0)
    DARKCYAN = (0, 127, 127)
    MAROON = (127, 0, 0)
    PURPLE = (127, 0, 127)
    OLIVE = (127, 127, 0)
    LIGHTGREY = (192, 192, 192)
    DARKGREY = (127, 127, 127)
    BLUE = (0, 0, 255)
    GREEN = (0, 255, 0)
    CYAN = (0, 255, 0)
    RED = (255, 0, 0)
    MAGENTA = (255, 0, 255)
    YELLOW = (255, 255, 0)
    WHITE = (255, 255, 255)
    ORANGE = (255, 165, 0)
    GREENYELLOW = (173, 255, 47)
    PINK = (255, 0, 255)



from __future__ import annotations

import contextlib
from typing import Any, Optional

from .config import AkariClientConfig, load_config
from .joint_manager import JointManager
from .m5stack_client import M5StackClient


class AkariClient:
    def __init__(self, config: Optional[AkariClientConfig] = None) -> None:
        self._stack = contextlib.ExitStack()

        self._config = config or load_config()
        self._joints = self._config.joint_manager.factory(self._stack)
        self._m5stack = self._config.m5stack.factory(self._stack)

    def __enter__(self) -> AkariClient:
        return self

    def __exit__(self, exc_type: Any, exc: Any, tb: Any) -> None:
        self.close()

    def close(self) -> None:
        self._stack.close()

    @property
    def joints(self) -> JointManager:
        return self._joints

    @property
    def m5stack(self) -> M5StackClient:
        return self._m5stack



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Optional, TypedDict

from .color import Color
from .position import Positions


class M5ComDict(TypedDict):
    """
    M5から取得する情報を格納するDict。

    :param din0: ヘッドのdin0ピンのデジタル入力。HiがTrue、LoがFalse。デフォルトHi。
    :param din1: ヘッドのdin1ピンのデジタル入力。HiがTrue、LoがFalse。デフォルトHi。
    :param ain0: ヘッドのain0ピンのアナログ入力。0-3.3Vを0-255の256段階にする。
    :param dout0: ヘッドのdout0ピンのデジタル出力。TrueでHi(3.3V)、FalseでLo(0V)。
    :param dout1: ヘッドのdout1ピンのデジタル出力。TrueでHi(3.3V)、FalseでLo(0V)。
    :param pwmout0: ヘッドのpwmout0ピンのPWM出力。0-255の256段階で0-3.3Vを出力可能。
    :param general0: 任意値を入力可能なエリア。通常時は使用しない。M5のソフトを書き換えることで自由に使用可能。
    :param general1: 任意値を入力可能なエリア。通常時は使用しない。M5のソフトを書き換えることで自由に使用可能。
    :param button_a: M5の前面のボタンAの入力。ボタンを押すとTrue、押していないとFalse。
    :param button_b: M5の前面のボタンBの入力。ボタンを押すとTrue、押していないとFalse。
    :param button_c: M5の前面のボタンCの入力。ボタンを押すとTrue、押していないとFalse。
    :param temperature: AKARI側面の環境センサモジュールの温度計測値。単位は℃。
    :param pressure: AKARI側面の環境センサモジュールの観測絶対圧力。単位はPa。
    :param brightness: AKARI側面の光センサユニットの観測明るさを4096段階で出力。0が最も明るい。
    :param time: データ取得した時間。M5のソフト起動時からの経過秒数を用いている。
    :param is_responce: akari_client()の各コマンド実行完了時のレスポンス。実行完了するとTrueを返す。

    """

    din0: bool
    din1: bool
    ain0: int
    dout0: bool
    dout1: bool
    pwmout0: int
    general0: float
    general1: float
    button_a: bool
    button_b: bool
    button_c: bool
    temperature: float
    pressure: float
    brightness: int
    time: float
    is_response: bool


class M5StackClient(ABC):
    @abstractmethod
    def set_dout(self, pin_id: int, value: bool, sync: bool = True) -> None:
        """ヘッド部GPIOピンのデジタル出力を設定する。

        Args:
            pin_id: pin番号。0でdout0、1でdout1を指定する。
            value: デジタル出力の値。Falseで0V,Trueで3.3V出力。
            sync: 同期実行の指定。Trueの場合M5側で実行完了するまで関数の終了待ちを行う。

        Example:
            >>> from akari_client import AkariClient
            >>> with AkariClient() as akari:
            >>>     m5 = akari.m5stack
            >>>     m5.set_dout(0, True)
            # dout0がTrueになる。

        """
        ...

    @abstractmethod
    def set_pwmout(self, pin_id: int, value: int, sync: bool = True) -> None:
        """ヘッド部GPIOピンのPWM出力を設定する。

        Args:
            pin_id: pin番号。デフォルトではpwmout0しかないため、0を指定すること。
            value: PWM出力の値。 0-255で指定し、0で0V、255で3.3Vを出力する。
            sync: 同期実行の指定。Trueの場合M5側で実行完了するまで関数の終了待ちを行う。

        Example:
            >>> from akari_client import AkariClient
            >>> with AkariClient() as akari:
            >>>     m5 = akari.m5stack
            >>>     m5.set_pwmout(0, 200)
            # pwmout0が200になる。

        """
        ...

    @abstractmethod
    def set_allout(
        self,
        *,
        dout0: Optional[bool] = None,
        dout1: Optional[bool] = None,
        pwmout0: Optional[int] = None,
        sync: bool = True,
    ) -> None:
        """ヘッド部GPIOピンの出力をまとめて設定する。

        Args:
            dout0: dout0の出力値。Falseで0V,Trueで3.3V出力。
            dout1: dout1の出力値。Falseで0V,Trueで3.3V出力。
            pwmout0: pwmout0の出力値。0-255で指定し、0で0V、255で3.3Vを出力する。
            sync: 同期実行の指定。Trueの場合M5側で実行完了するまで関数の終了待ちを行う。

        Example:
            >>> from akari_client import AkariClient
            >>> with AkariClient() as akari:
            >>>     m5 = akari.m5stack
            >>>     m5.set_allout(dout0=False, dout1=True, pwmout0=100)
            # dout0がFalse, dout1がTrue, pwmout0が100になる。

        """
        ...

    @abstractmethod
    def reset_allout(self, sync: bool = True) -> None:
        """ヘッド部GPIOピンの出力を初期値にリセットする。
        dout0、dout1はFalse、pwmout0は0となる。

        Args:
            sync: 同期実行の指定。Trueの場合M5側で実行完了するまで関数の終了待ちを行う。

        Example:
            >>> from akari_client import AkariClient
            >>> with AkariClient() as akari:
            >>>     m5 = akari.m5stack
            >>>     m5.reset_allout()
            # dout0がFalse, dout1がFalse, pwmout0が0になる。

        """
        ...

    @abstractmethod
    def set_display_color(
        self,
        color: Color,
        sync: bool = True,
    ) -> None:
        """ボディー部M5のディスプレイ背景色を変更する。

        Args:
            color: 背景色を指定。色は ``color.Colors`` から色名を引用する、もしくはRGBの数値指定も可能。
            sync: 同期実行の指定。Trueの場合M5側で実行完了するまで関数の終了待ちを行う。

        Example:
            >>> from akari_client import AkariClient
            >>> with AkariClient() as akari:
            >>>     m5 = akari.m5stack
            >>>     color = Colors.WHITE
            >>>     m5.set_display_color(color)
            # 画面色が白になる。
            >>>     color = Color(red=0, green=100, blue=200)
            >>>     m5.set_display_color(color)
            # 画面色が指定したRGB値になる。

        """
        ...

    @abstractmethod
    def set_display_text(
        self,
        text: str,
        pos_x: int = Positions.CENTER,
        pos_y: int = Positions.CENTER,
        size: int = 5,
        text_color: Optional[Color] = None,
        back_color: Optional[Color] = None,
        refresh: bool = True,
        sync: bool = True,
    ) -> None:
        """ボディー部M5のディスプレイに文字を表示する。

        Args:
            text: 表示する文字列。
            pos_x: x方向の描画位置ピクセルを0-320で指定。左端が0。 ``position.Positions`` を用いた位置指定も可能。デフォルト値は中央揃え。
            pos_y: y方向の描画位置ピクセルを0-240で指定。上端が0。 ``position.Positions`` を用いた位置指定も可能。デフォルト値は中央揃え。
            size: 文字サイズを1-11の11段階で指定。デフォルト値は5。
            text_color: 文字色を指定。色は ``color.Colors`` から色名を引用する、もしくはRGBの数値指定も可能。値を指定しない場合、前回値を引き継ぐ。
            back_color: 背景色を指定。色は ``color.Colors`` から色名を引用する、もしくはRGBの数値指定も可能。値を指定しない場合、画面全体の背景色に合わせる。
            refresh: trueの場合画面全体を現在の背景色で更新する。falseの場合は現在の表示を維持しつつ、文字を描画する範囲のみ更新する。デフォルト値は背景更新あり。
            sync: 同期実行の指定。Trueの場合M5側で実行完了するまで関数の終了待ちを行う。

        Example:
            >>> from akari_client import AkariClient
            >>> from akari_client.color import Colors
            >>> from akari_client.position import Positions
            >>> with AkariClient() as akari:
            >>>    text = "AKARI"
            >>>    pos_x = Positions.LEFT
            >>>    pos_y = Positions.TOP
            >>>    size = 7
            >>>    text_color = Colors.WHITE
            >>>    back_color = Colors.BLACK
            >>>    refresh = True
            >>>    m5.set_display_text(
            ...        text, pos_x, pos_y, size, text_color, back_color, refresh
            ...    )
            # 画面に「AKARI」と表示される。

        """
        ...

    @abstractmethod
    def set_display_image(
        self,
        filepath: str,
        pos_x: int = Positions.CENTER,
        pos_y: int = Positions.CENTER,
        scale: float = -1.0,
        sync: bool = True,
    ) -> None:
        """ボディー部M5のディスプレイにM5のSDカード内の画像を表示する。

        Args:
            filepath: M5のSDカード内のファイルパス。(例;"image/hoge.jpg")
            pos_x: x方向の描画位置ピクセルを0-320で指定。左端が0。 ``position.Positions`` を用いた位置指定も可能。デフォルト値は中央揃え。
            pos_y: y方向の描画位置ピクセルを0-240で指定。上端が0。 ``position.Positions`` を用いた位置指定も可能。デフォルト値は中央揃え。
            scale: 画像の拡大縮小倍率を指定。1.0で等倍表示。マイナスの値を入れた場合、画面サイズに合わせて自動でサイズ調整される。
                   デフォルト値は自動サイズ調整となっている。
            sync: 同期実行の指定。Trueの場合M5側で実行完了するまで関数の終了待ちを行う。

        Example:
            >>> from akari_client import AkariClient
            >>> from akari_client.position import Positions
            >>> with AkariClient() as akari:
            >>>    filepath = "/jpg/logo320.jpg"
            >>>    pos_x = Positions.LEFT
            >>>    pos_y = Positions.TOP
            >>>    scale = 0.75
            >>>    m5.set_display_image(filepath, pos_x, pos_y, scale)
            # 画面に"/jpg/logo320.jpg"の画像が表示される。

        """
        ...

    @abstractmethod
    def reset_m5(self) -> None:
        """M5をリセットして再起動する。

        Example:
            >>> from akari_client import AkariClient
            >>> with AkariClient() as akari:
            >>>    m5.reset_m5()
            # M5が再起動する。

        """
        ...

    @abstractmethod
    def get(self) -> M5ComDict:
        """M5から環境センサ、ヘッドGPIOの入力値、M5のボタンの状態を取得する。

        Returns:
            M5から取得した各データを格納したDict。

        Example:
            >>> from akari_client import AkariClient
            >>> with AkariClient() as akari:
            >>>    data = m5.get()
            >>>    print(data["temperature"])
            31.675

        """
        ...



from setuptools import find_packages, setup

setup(
    name="akari-client",
    version="0.3.2",
    packages=find_packages(exclude=["tests"]),
    description="Akari Python package",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="akari",
    author_email="akari.tmc@gmail.com",
    license="Apache License 2.0",
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: POSIX",
        "Operating System :: Unix",
    ],
    url="https://github.com/AkariGroup/akari_software",
    install_requires=[
        "dynamixel_sdk",
        "pydantic>=1.5,<2.0",
    ],
    package_data={"akari_client": ["py.typed"]},
    extras_require={
        "grpc": [
            "akari-proto>=0.3.0,<0.4.0",
            "grpcio==1.44.0",
            "protobuf==3.19.3",
        ],
        "depthai": [
            "matplotlib",
            "depthai",
            "opencv-python",
            "blobconverter",
        ],
    },
)



import abc
from typing import NamedTuple


class PositionLimit(NamedTuple):
    min: float
    max: float


class RevoluteJointController(abc.ABC):
    @property
    @abc.abstractmethod
    def joint_name(self) -> str:
        ...

    @abc.abstractmethod
    def get_servo_enabled(self) -> bool:
        """サーボの有効無効状態を取得する。

        Returns:
            サーボ有効なら``True``、無効なら``False``

        """
        ...

    @abc.abstractmethod
    def set_servo_enabled(self, enabled: bool) -> None:
        """サーボの有効無効状態を設定する。

        Args:
            enabled : ``True`` であればサーボを有効にする

        """
        ...

    @abc.abstractmethod
    def get_position_limit(self) -> PositionLimit:
        """Positionの上限値と下限値を取得する。

        Returns:
            現在角度の下限値、上限値 [rad]
        """
        ...

    @abc.abstractmethod
    def set_profile_acceleration(self, rad_per_sec2: float) -> None:
        """サーボの目標加速度を設定する。

        Args:
            rad_per_sec2: 加速度 [rad/s^2]

        """
        ...

    @abc.abstractmethod
    def get_profile_acceleration(self) -> float:
        """Profile Acceleration を取得する。

        Returns:
            加速度 [rad/s^2]
        """
        ...

    @abc.abstractmethod
    def set_profile_velocity(self, rad_per_sec: float) -> None:
        """サーボの目標速度を設定する。

        Args:
            rad_per_sec: 速度 [rad/s]

        """
        ...

    @abc.abstractmethod
    def get_profile_velocity(self) -> float:
        """Profile Velocity を設定する

        Args:
            rad_per_sec: 速度 [rad/s]
        """
        ...

    @abc.abstractmethod
    def set_goal_position(self, rad: float) -> None:
        """サーボの目標角度を設定する。
        ここで設定した値まで移動する。

        Args:
            rad: 目標角度 [rad]

        """
        ...

    @abc.abstractmethod
    def get_present_position(self) -> float:
        """サーボの現在角度を取得する。

        Returns:
            現在角度 [rad]

        """
        ...

    @abc.abstractmethod
    def get_moving_state(self) -> bool:
        """モーターが動作中かどうか判定する。

        Returns:
            現在のモーター状態。
        """
        ...



import json
from typing import Dict, Iterator, Optional, cast

import grpc
from akari_client.color import Color
from akari_client.position import Positions
from akari_proto import m5stack_pb2
from akari_proto.grpc.error import deserialize_error
from akari_proto.m5stack_pb2_grpc import M5StackServiceStub
from google.protobuf.empty_pb2 import Empty

from ..m5stack_client import M5ComDict, M5StackClient
from ._error import serializer


def _as_proto_color(color: Optional[Color]) -> m5stack_pb2.Color:
    if color is None:
        return m5stack_pb2.Color(red=-1, green=-1, blue=-1)

    return m5stack_pb2.Color(
        red=color.red,
        green=color.green,
        blue=color.blue,
    )


class GrpcM5StackClient(M5StackClient):
    def __init__(self, channel: grpc.Channel) -> None:
        self._stub = M5StackServiceStub(channel)

    @deserialize_error(serializer)
    def set_dout(self, pin_id: int, value: bool, sync: bool = True) -> None:
        binary_pins: Dict[str, bool] = {}
        if pin_id == 0:
            binary_pins["dout0"] = value
        elif pin_id == 1:
            binary_pins["dout1"] = value
        else:
            raise ValueError(f"Out of range pin_id: {pin_id}")

        request = m5stack_pb2.SetPinOutRequest(
            binary_pins=binary_pins,
            sync=sync,
        )
        self._stub.SetPinOut(request)

    @deserialize_error(serializer)
    def set_pwmout(self, pin_id: int, value: int, sync: bool = True) -> None:
        int_pins: Dict[str, int] = {}
        if pin_id == 0:
            int_pins["pwmout0"] = value
        else:
            raise ValueError(f"Out of range pin_id: {pin_id}")

        request = m5stack_pb2.SetPinOutRequest(
            int_pins=int_pins,
            sync=sync,
        )
        self._stub.SetPinOut(request)

    @deserialize_error(serializer)
    def set_allout(
        self,
        *,
        dout0: Optional[bool] = None,
        dout1: Optional[bool] = None,
        pwmout0: Optional[int] = None,
        sync: bool = True,
    ) -> None:
        binary_pins: Dict[str, bool] = {}
        int_pins: Dict[str, int] = {}

        if dout0 is not None:
            binary_pins["dout0"] = dout0
        if dout1 is not None:
            binary_pins["dout1"] = dout1
        if pwmout0 is not None:
            int_pins["pwmout0"] = pwmout0

        request = m5stack_pb2.SetPinOutRequest(
            binary_pins=binary_pins,
            int_pins=int_pins,
            sync=sync,
        )
        self._stub.SetPinOut(request)

    @deserialize_error(serializer)
    def reset_allout(self, sync: bool = True) -> None:
        self._stub.ResetPinOut(m5stack_pb2.ResetPinOutRequest(sync=sync))

    @deserialize_error(serializer)
    def set_display_color(self, color: Color, sync: bool = True) -> None:
        request = m5stack_pb2.SetDisplayColorRequest(
            color=_as_proto_color(color),
            sync=sync,
        )
        self._stub.SetDisplayColor(request)

    @deserialize_error(serializer)
    def set_display_text(
        self,
        text: str,
        pos_x: int = Positions.CENTER,
        pos_y: int = Positions.CENTER,
        size: int = 5,
        text_color: Optional[Color] = None,
        back_color: Optional[Color] = None,
        refresh: bool = True,
        sync: bool = True,
    ) -> None:
        request = m5stack_pb2.SetDisplayTextRequest(
            text=text,
            pos_x=pos_x,
            pos_y=pos_y,
            size=size,
            text_color=_as_proto_color(text_color),
            bg_color=_as_proto_color(back_color),
            refresh=refresh,
            sync=sync,
        )
        self._stub.SetDisplayText(request)

    @deserialize_error(serializer)
    def set_display_image(
        self,
        filepath: str,
        pos_x: int = Positions.CENTER,
        pos_y: int = Positions.CENTER,
        scale: float = -1.0,
        sync: bool = True,
    ) -> None:
        request = m5stack_pb2.SetDisplayImageRequest(
            path=filepath,
            pos_x=pos_x,
            pos_y=pos_y,
            scale=scale,
            sync=sync,
        )
        self._stub.SetDisplayImage(request)

    @deserialize_error(serializer)
    def reset_m5(self) -> None:
        self._stub.Reset(Empty())

    @deserialize_error(serializer)
    def get(self) -> M5ComDict:
        data: m5stack_pb2.M5StackStatus = self._stub.Get(Empty())
        return cast(M5ComDict, json.loads(data.status_json))

    @deserialize_error(serializer)
    def get_stream(self) -> Iterator[M5ComDict]:
        stream = self._stub.GetStream(Empty())
        for r in stream:
            data: m5stack_pb2.M5StackStatus = r
            yield cast(M5ComDict, json.loads(data.status_json))



from __future__ import annotations

import enum
import time
from typing import Dict, Iterator, List, Optional, Sequence, Tuple, TypeVar

from .joint_controller import PositionLimit, RevoluteJointController

TValue = TypeVar("TValue")


class AkariJoint(str, enum.Enum):
    """
    AKARIの関節名の一覧。
    """

    PAN = "pan"
    TILT = "tilt"


class JointManager:
    def __init__(self, joint_controllers: Sequence[RevoluteJointController]) -> None:
        """Akariの関節制御コントローラ"""

        self._joints: Dict[str, RevoluteJointController] = {}
        for j in joint_controllers:
            self._joints[j.joint_name] = j

    @property
    def joint_controllers(self) -> Sequence[RevoluteJointController]:
        return list(self._joints.values())

    def get_joint_names(self) -> List[str]:
        """関節名を取得する。

        Returns:
            AKARIの全関節名のlist

        """
        return list(self._joints.keys())

    @property
    def pan_joint(self) -> RevoluteJointController:
        return self._joints[AkariJoint.PAN]

    @property
    def tilt_joint(self) -> RevoluteJointController:
        return self._joints[AkariJoint.TILT]

    def _iter_joint_value_pairs(
        self,
        pan: Optional[TValue] = None,
        tilt: Optional[TValue] = None,
        **kwargs: TValue,
    ) -> Iterator[Tuple[RevoluteJointController, TValue]]:
        values: Dict[str, TValue] = {}
        if pan is not None:
            values[AkariJoint.PAN] = pan
        if tilt is not None:
            values[AkariJoint.TILT] = tilt
        values.update(kwargs)

        for joint_name, value in values.items():
            controller = self._joints[joint_name]
            yield controller, value

    def get_joint_limits(self) -> Dict[str, PositionLimit]:
        """サーボの位置リミットを取得する。

        Returns:
            関節名と位置リミット(min,max)[rad]のDict

        Example:
            >>> from akari_client import AkariClient
            >>> akari = AkariClient()
            >>> joints = akari.joints
            >>> print(joints.get_joint_limits())
            # 各軸のリミット値が出力される

        """
        ret: Dict[str, PositionLimit] = {}
        for joint_name, controller in self._joints.items():
            ret[joint_name] = controller.get_position_limit()
        return ret

    def set_joint_accelerations(
        self,
        *,
        pan: Optional[float] = None,
        tilt: Optional[float] = None,
        **kwargs: float,
    ) -> None:
        """サーボの目標加速度を設定する。

        Args:
            pan: pan軸の加速度 [rad/s^2]
            tilt: tilt軸の加速度 [rad/s^2]

        Example:
            >>> from akari_client import AkariClient
            >>> akari = AkariClient()
            >>> joints = akari.joints
            >>> joints.set_joint_accelerations(pan=10, tilt=8)
            # pan軸の目標加速度が10rad/s^2, tilt軸の目標加速度が8rad/s^2

        """
        for joint, value in self._iter_joint_value_pairs(pan, tilt, **kwargs):
            joint.set_profile_acceleration(value)

    def get_joint_accelerations(self) -> Dict[str, float]:
        """サーボの目標加速度を取得する。

        Returns:
            関節名と目標加速度[rad/s^2]のDict

        Example:
            >>> from akari_client import AkariClient
            >>> akari = AkariClient()
            >>> joints = akari.joints
            >>> print(joints.get_joint_accelerations())
            # 各軸の目標加速度値が出力される

        """
        ret: Dict[str, float] = {}
        for joint_name, controller in self._joints.items():
            ret[joint_name] = controller.get_profile_acceleration()
        return ret

    def set_joint_velocities(
        self,
        *,
        pan: Optional[float] = None,
        tilt: Optional[float] = None,
        **kwargs: float,
    ) -> None:
        """サーボの目標速度を設定する。

        Args:
            pan: pan軸の速度 [rad/s]
            tilt: tilt軸の速度 [rad/s]

        Example:
            >>> from akari_client import AkariClient
            >>> akari = AkariClient()
            >>> joints = akari.joints
            >>> joints.set_joint_velocities(pan=10, tilt=8)
            # pan軸の目標加速度が10rad/s, tilt軸の目標加速度が8rad/s

        """
        for joint, value in self._iter_joint_value_pairs(pan, tilt, **kwargs):
            joint.set_profile_velocity(value)

    def get_joint_velocities(self) -> Dict[str, float]:
        """サーボの目標速度を取得する。

        Returns:
            関節名と目標速度[rad/s]のDict

        Example:
            >>> from akari_client import AkariClient
            >>> akari = AkariClient()
            >>> joints = akari.joints
            >>> print(joints.get_joint_velocities())
            # 各軸の目標速度値が出力される

        """
        ret: Dict[str, float] = {}
        for joint_name, controller in self._joints.items():
            ret[joint_name] = controller.get_profile_velocity()
        return ret

    def move_joint_positions(
        self,
        *,
        sync: bool = False,
        pan: Optional[float] = None,
        tilt: Optional[float] = None,
        **kwargs: float,
    ) -> None:
        """サーボの目標角度を設定する。
        ここで設定した値まで移動する。

        Args:
            sync: Trueにすると、サーボの移動が完了するまで関数の終了を待機する。デフォルト値はFalse。
            pan: pan軸の目標角度 [rad]
            tilt: tilt軸の目標角度 [rad]

        Example:
            >>> from akari_client import AkariClient
            >>> akari = AkariClient()
            >>> joints = akari.joints
            >>> joints.move_joint_positions(pan=0.4,tilt=-0.2)
            # pan軸が0.4rad, tilt軸の目標加速度が-0.2radの位置へ移動

        """
        for joint, position in self._iter_joint_value_pairs(pan, tilt, **kwargs):
            joint.set_goal_position(position)
        if sync:
            while True:
                for joint, position in self._iter_joint_value_pairs(
                    pan, tilt, **kwargs
                ):
                    if not joint.get_moving_state():
                        break
                else:
                    break
                time.sleep(0.01)

    def get_joint_positions(self) -> Dict[str, float]:
        """サーボの現在角度を取得する。

        Returns:
            サーボ名と現在角度[rad]のdict

        Example:
            >>> from akari_client import AkariClient
            >>> akari = AkariClient()
            >>> joints = akari.joints
            >>> print(joints.get_joint_positions())
            # 各軸の現在角度[rad]が出力される

        """
        ret: Dict[str, float] = {}
        for joint_name, controller in self._joints.items():
            ret[joint_name] = controller.get_present_position()

        return ret

    def set_servo_enabled(
        self,
        *,
        pan: Optional[bool] = None,
        tilt: Optional[bool] = None,
        **kwargs: bool,
    ) -> None:
        """サーボの有効無効状態を設定する。

        Args:
            pan: pan軸のサーボ有効無効の設定。``True`` であればサーボを有効にする。
            tilt: tilt軸のサーボ有効無効の設定。``True`` であればサーボを有効にする。

        Example:
            >>> from akari_client import AkariClient
            >>> akari = AkariClient()
            >>> joints = akari.joints
            >>> joints.set_servo_enabled(pan=True,tilt=False)
            # panのサーボを有効、tiltのサーボを無効に設定

        """
        for joint, value in self._iter_joint_value_pairs(pan, tilt, **kwargs):
            joint.set_servo_enabled(value)

    def enable_all_servo(self) -> None:
        """全サーボを有効状態に設定する。

        Example:
            >>> from akari_client import AkariClient
            >>> akari = AkariClient()
            >>> joints = akari.joints
            >>> joints.enable_all_servo()
            # pan、tiltのサーボ両方を有効に設定

        """
        for joint in self._joints.values():
            joint.set_servo_enabled(True)

    def disable_all_servo(self) -> None:
        """全サーボを無効状態に設定する。

        Example:
            >>> from akari_client import AkariClient
            >>> akari = AkariClient()
            >>> joints = akari.joints
            >>> joints.disable_all_servo()
            # pan、tiltのサーボ両方を無効に設定

        """
        for joint in self._joints.values():
            joint.set_servo_enabled(False)
